#########################################################################################
# Chatologica - FileLocks.pm - reading file and directories tools
# All rights reserved (c) 2000; http://www.chatologica.com/
#########################################################################################

package DBPUB1::STD::FileLocks;

use strict;				# use strict pragma
use DBPUB1::STD::Templater;             # load standard routines related to templates



sub get_lock                                    # get a file lock
{
    my($file) = @_;
    $file =~ s{[/\\]}{][}g;                     # encode slashes
    $file =~ s{\s}{_}g;                         # encode whitespaces
    $file = "locks/$file";                      # make as file name under locks directory
    my $fh = '';                                # file handler
    my $problem = '';				# problem in file lock getting?
    my $not_supported = '';			# is flock supported
    ($fh,$problem) = &file_open("$file",'>',1,'LOCKER'); # open and lock it
    if($problem) {  				# flock() may not be supported on all platforms
        close $fh;
        unlink $file;                           # remove the file for unsuccessful flock()
        ($fh, $file) = &custom_lock($fh, $file);# flock not implemented or not successful
    };
    return ($fh, $file);
}



sub release_lock                        # release a lock on file
{
    my ($fh, $file) = @_;
    if((defined $fh) && $fh) {
        close($fh);                     # close file if we have opened filehandle yet
    };
    if((defined $file) && ($file =~ /\.flock$/)) {
    	unlink ($file);                	# delete the lock file
    };
}



sub file_open					# opens and locks a file exclusively
{                                      		# filename must start
    my($file,$direction,$exclusive,$handler_name) = @_;
    my $fh = '';				# a file handler
    $handler_name = $handler_name || 'File';	# set a default handler name if not provided
    if($ENV{'FCGI_ROLE'} || $ENV{'MOD_PERL'} || '') {   # under mod_perl/fast_cgi use IO::File to open files
        eval("use IO::File");                   # load only if requested because is CPU ineffective to use it
        $fh = new IO::File;                     # standard cgi-bin mode. Now making new file handler
        open($fh,"$direction$file") || &DieMsg("Can't open file <I>'$file'</I>: $!");
    } else {
	eval qq~
        open($handler_name,"\$direction\$file") || &DieMsg("Can't open file <I>'\$file'</I>: $!");
        \$fh = \*$handler_name;
	~;
    };
    my $problem = '';				    	# no problems for now
    if((defined $exclusive) && $exclusive) {
    	local $^W = 0;                              	# Warnings OFF
    	my $die_handler = $SIG{__DIE__};		# remember the old handler
    	local $SIG{__DIE__} = undef;                	# to avoid error message if flock is unimplemented
	$problem = (!eval("flock(\$fh,2)")) || $@;
    	local $SIG{__DIE__} = $die_handler; 		# restore original handler    	
    };
    return ($fh,$problem);
}



sub custom_lock                         # we run this subroutine if flock() is not implemented
{
    my ($fh, $file) = @_;
    $file .= '.flock';                          # make custom lock filename
    my $timeout = 120;                        	# 120 seconds max cgi script critical section life
    my $end_time = time+15;                     # do not wait more than 15 seconds for getting the lock 
    while (-e $file && (time <= $end_time)) {   # wait to get the lock file deleted
        select(undef, undef, undef, 0.25);      # sleep 250 miliseconds
    };
    if( ( (time > $end_time) && (((stat($file))[9]+$timeout)<time)) || (time < $end_time) ) {  # we can write a lock file now
        open(LOCK,">$file") || &DieMsg("Can't write <I>'$file'</I>: $!");
        print LOCK $$;                          # write the process id in the lock file
        close LOCK;                             # save this Process ID in the lock file
        sleep(1);                               # wait 1 second
        open(LOCK,"<$file") || &DieMsg("Can't read <I>'$file'</I>: $!");
        my $pid = <LOCK>;                       # read again the Process ID from file and compare with the current
        close LOCK;
        if($pid eq $$) {
            return ('',$file);                  # SUCCESSFULLY GOTTEN LOCK ON FILE!!!
        };
    };
    &DieMsg("Can't get lock on file <i>'$file'</i> within reasonable time. May be the system is too busy now - try again!");
};



sub import {			# this function is called automatically by use Module; statement
    my $pkg = shift;
    my $callpkg = caller(0);	# import these symbols into the caller package:
    no strict 'refs';
    foreach $_ (qw(get_lock release_lock file_open)) {
    	*{"${callpkg}::$_"} = \&{"${pkg}::$_"};
    };
}



1; # must return TRUE

