############################################################################################
# Chatologica - FlatDB.pm - process flat-text databases
# All rights reserved (c) 2000; http://www.chatologica.com/
############################################################################################

package DBPUB1::STD::FlatDB;

use strict;
use DBPUB1::STD::Templater;                     # load standard routines related to templates


sub new {                                       # a standard object constructor
    my ($this, @params) = @_;
    my $class = ref($this) || $this;
    my $self = {};
    bless $self, $class;
    $self->initialize(@params);
    return $self;
}



sub initialize
{
    my ($self,$db_file,$structure) = @_;
    $structure = $structure || [];              # init structure
    $self->{'file'} = $db_file || '';		# file name of the database
    $self->{'structure'} = $structure;          # ref. to list of ordered field names
    $self->{'key'} = $$structure[0] || '';      # key field name
    $self->{'content'} = ();			# init database content
    $self->{'keys'} = ();			# init database key list
};



sub load					# load the database in RAM
{
    my ($self) = @_;
    my ($row, @row) = ();			# the current row as string or array
    $self->{'content'} = ();			# init database content
    $self->{'keys'} = ();			# init database key list
    open(FDB,"<$self->{'file'}") || &DieMsg("Can't read file <i>'$self->{'file'}'</i>: $!");
    while((defined($row = <FDB>)) && $row){
	chomp $row;				# remove new line at the end of row
        @row = split(/\|/,$row);		# split by field separator
	if (@row) {				# valid row
	    $row[0] = $row[0] || '';		# allow empty ids
	    my %row = (); 			# allocate memory for a hash
            foreach (0..$#{$self->{'structure'}}) { # parse a database record and build a hash
                $row{$self->{'structure'}[$_]} = $row[$_] || '';
            };
	    push @{$self->{'keys'}}, $row[0];	# add this key to the array of keys
	    $self->{'content'}{$row[0]} = \%row;# build hash of hashes
	};
    };
};



sub add_row							# add/update a new database row
{
    my ($self, $row) = @_;
    if ((defined $row) && (defined $$row{$self->{'key'}})) {	# is new row legal
	if(!($self->{'content'}{$$row{$self->{'key'}}})) { 	# it is a new record
            unshift @{$self->{'keys'}}, $$row{$self->{'key'}};  # add to the list of fields
    	};
        foreach (0..$#{$self->{'structure'}}) { # parse a database record and build a hash
            $$row{$self->{'structure'}[$_]} = $$row{$self->{'structure'}[$_]} || '';
        };
    	$self->{'content'}{$$row{$self->{'key'}}} = $row;   	# update the hash for this new record
    };
}



sub delete_rows					# delete rows for which $criteria expression is 1
{
    my ($self, $criteria) = @_;
    my ($key, @found) = ();
    local $^W = 0;                                      # disable warnings
    foreach (0..$#{$self->{'keys'}}) {         		# browse database hash
   	$key = $self->{'keys'}[$_];			# make the current record id
        if(exists $self->{'content'}{$key}) {           # test only rows with valid keys
	    if(&$criteria($self->{'content'}{$key})){	# test the criteria
	    	delete $self->{'content'}{$key};	# delete this row
                $self->{'keys'}[$_] = undef;            # undef this key
	    };
    	};
    };
}



sub search                 # search for rows which return 1 for $criteria expression
{
    my ($self, $criteria, $id, $start_index, $length, $sort_by, $descending, $numerical) = @_;
    my ($key, @found, $total, %by_field) = ();
    if ((defined $id) && (defined $self->{'content'}{$id})) {
	push @found, $self->{'content'}{$id};		# find a row by a given key
        return (\@found,1);
    };
    local $^W = 0;                                      # disable warnings
    foreach $key (@{$self->{'keys'}}) {                 # browse database hash
        if((defined $key) and (exists $self->{'content'}{$key})) { # test only rows with valid keys
            if(&{$criteria}($self->{'content'}{$key})){ # test the criteria
		push @found, $self->{'content'}{$key};	# add this row to the list of results
	    };
	};
    };
    $sort_by = $sort_by || ''; $descending = $descending || '';	# init
    if($sort_by) {					# have to sort results
	if(!$descending) {				# in ascending order
            if($numerical) {
                @found = sort {$$a{$sort_by} <=> $$b{$sort_by}} @found;
            } else {
                @found = sort {$$a{$sort_by} cmp $$b{$sort_by}} @found;
            };
	} else {
            if($numerical) {
                @found = sort {$$b{$sort_by} <=> $$a{$sort_by}} @found;
            } else {
                @found = sort {$$b{$sort_by} cmp $$a{$sort_by}} @found;
            };
	};
    };
    $total = $#found + 1;				# total number of results found
    $length = $length || 50;
    if($total && ($start_index > $total)) { 		# if start index is too high
        $start_index = 1 + $length * (int(($total-1)/$length));
    };
    if($start_index+$length-1 > $total) {		# fixing the length of results requested
        $length = 1 + $total - $start_index;
    };
    if($length < 0) {					# only only non-negative lengths
	$length = 0;
    };
    @found = splice(@found, $start_index-1, $length);   # give only these numbers from the results
    return (\@found, $total);           		# found results as list of hashes
}



sub save						# save this database file
{
    my ($self) = @_;
    my ($key, $row, $field, $field_value, @values, %saved) = ();
    open(FDB,">$self->{'file'}") || &DieMsg("Can't write to file <i>'$self->{'file'}</i>: $!");
    foreach $key (@{$self->{'keys'}}) {                 # browse database hash
        if((defined $key) and (exists $self->{'content'}{$key})         # if not-deleted
            and !exists $saved{$key}) {                                 # and non-saved yet row
            @values = ();                               # init values
            foreach $field (@{$self->{'structure'}}) {	# prepare a record string
            	$field_value = $self->{'content'}{$key}{$field} || '';
	    	$field_value =~ s/\|/:/g;		# translate the field separator to :
            	push @values, $field_value;
            };
            $row = join('|',@values) . "\n";          # make the record as a string
            $saved{$key} = 1;
	    print FDB "$row";
        };
    };
    close FDB;                                	# Important!!!
    chmod 0666, "$self->{'file'}";		# make file writeable
};



1; # must return TRUE

